import os
import time
import pstats
import typing
import logging
import cProfile
from importlib import resources
from functools import lru_cache
from typing import Literal, Callable, TYPE_CHECKING, TypeGuard

import cv2
from cv2.typing import MatLike

if TYPE_CHECKING:
    from kotonebot.client.protocol import Device
    from kotonebot.backend.color import HsvColor
from .core import Image

logger = logging.getLogger(__name__)



Rect = typing.Sequence[int]
"""左上X, 左上Y, 宽度, 高度"""

def is_rect(rect: typing.Any) -> TypeGuard[Rect]:
    return isinstance(rect, typing.Sequence) and len(rect) == 4 and all(isinstance(i, int) for i in rect)

def crop(img: MatLike, /, x1: float = 0, y1: float = 0, x2: float = 1, y2: float = 1) -> MatLike:
    """
    按比例裁剪图像。

    :param img: 图像
    :param x1: 裁剪区域左上角相对X坐标。范围 [0, 1]，默认为 0
    :param y1: 裁剪区域左上角相对Y坐标。范围 [0, 1]，默认为 0
    :param x2: 裁剪区域右下角相对X坐标。范围 [0, 1]，默认为 1
    :param y2: 裁剪区域右下角相对Y坐标。范围 [0, 1]，默认为 1
    """
    h, w = img.shape[:2]
    x1_px = int(w * x1)
    y1_px = int(h * y1) 
    x2_px = int(w * x2)
    y2_px = int(h * y2)
    return img[y1_px:y2_px, x1_px:x2_px]

def crop_rect(img: MatLike, rect: Rect) -> MatLike:
    """
    按范围裁剪图像。

    :param img: 图像
    :param rect: 裁剪区域。
    """
    x, y, w, h = rect
    return img[y:y+h, x:x+w]

class DeviceHookContextManager:
    def __init__(
        self,
        device: 'Device',
        *,
        screenshot_hook_before: Callable[[], MatLike|None] | None = None,
        screenshot_hook_after: Callable[[MatLike], MatLike] | None = None,
        click_hook_before: Callable[[int, int], tuple[int, int]] | None = None,
    ):
        self.device = device
        self.screenshot_hook_before = screenshot_hook_before
        self.screenshot_hook_after = screenshot_hook_after
        self.click_hook_before = click_hook_before

        self.old_screenshot_hook_before = self.device.screenshot_hook_before
        self.old_screenshot_hook_after = self.device.screenshot_hook_after
    
    def __enter__(self):
        if self.screenshot_hook_before is not None:
            self.device.screenshot_hook_before = self.screenshot_hook_before
        if self.screenshot_hook_after is not None:
            self.device.screenshot_hook_after = self.screenshot_hook_after
        if self.click_hook_before is not None:
            self.device.click_hooks_before.append(self.click_hook_before)
        return self.device
    
    def __exit__(self, exc_type, exc_value, traceback):
        self.device.screenshot_hook_before = self.old_screenshot_hook_before
        self.device.screenshot_hook_after = self.old_screenshot_hook_after
        if self.click_hook_before is not None:
            self.device.click_hooks_before.remove(self.click_hook_before)

def cropped(
    device: 'Device',
    x1: float = 0,
    y1: float = 0,
    x2: float = 1,
    y2: float = 1,
) -> DeviceHookContextManager:
    """
    Hook 设备截图与点击操作，将截图裁剪为指定区域，并调整点击坐标。

    在进行 OCR 识别或模板匹配时，可以先使用此函数缩小图像，加快速度。

    :param device: 设备对象
    :param x1: 裁剪区域左上角相对X坐标。范围 [0, 1]，默认为 0
    :param y1: 裁剪区域左上角相对Y坐标。范围 [0, 1]，默认为 0
    :param x2: 裁剪区域右下角相对X坐标。范围 [0, 1]，默认为 1
    :param y2: 裁剪区域右下角相对Y坐标。范围 [0, 1]，默认为 1
    """
    def _screenshot_hook(img: MatLike) -> MatLike:
        return crop(img, x1, y1, x2, y2)
    def _click_hook(x: int, y: int) -> tuple[int, int]:
        w, h = device.screen_size
        x_px = int(x1 * w + x)
        y_px = int(y1 * h + y)
        return x_px, y_px
    return DeviceHookContextManager(
        device,
        screenshot_hook_after=_screenshot_hook,
        click_hook_before=_click_hook,
    )

def grayscaled(img: MatLike | str | Image) -> MatLike:
    if isinstance(img, str):
        img = cv2.imread(img)
    elif isinstance(img, Image):
        img = img.data
    return cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

@lru_cache
def grayscale_cached(img: MatLike | str) -> MatLike:
    return grayscaled(img)

def until(
    condition: Callable[[], bool],
    timeout: float=60,
    interval: float=0.5,
    critical: bool=False
) -> bool:
    """
    等待条件成立，如果条件不成立，则返回 False 或抛出异常。

    :param condition: 条件函数。
    :param timeout: 等待时间，单位为秒。
    :param interval: 检查条件的时间间隔，单位为秒。
    :param critical: 如果条件不成立，是否抛出异常。
    """
    start = time.time()
    while not condition():
        if time.time() - start > timeout:
            if critical:
                raise TimeoutError(f"Timeout while waiting for condition {condition.__name__}.")
            return False
        time.sleep(interval)
    return True


class AdaptiveWait:
    """
    自适应延时。延迟时间会随着时间逐渐增加，直到达到最大延迟时间。
    """
    def __init__(
        self,
        base_interval: float = 0.5,
        max_interval: float = 10,
        *,
        timeout: float = -1,
        timeout_message: str = "Timeout",
        factor: float = 1.15,
    ):
        self.base_interval = base_interval
        self.max_interval = max_interval
        self.interval = base_interval
        self.factor = factor
        self.timeout = timeout
        self.start_time: float | None = time.time()
        self.timeout_message = timeout_message

    def __enter__(self):
        self.start_time = time.time()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.reset()

    def __call__(self):
        from .context import sleep
        if self.start_time is None:
            self.start_time = time.time()
        sleep(self.interval)
        self.interval = min(self.interval * self.factor, self.max_interval)
        if self.timeout > 0 and time.time() - self.start_time > self.timeout:
            raise TimeoutError(self.timeout_message)

    def reset(self):
        self.interval = self.base_interval
        self.start_time = None

class Countdown:
    def __init__(self, sec: float):
        self.seconds = sec
        self.start_time: float | None = None

    def __str__(self):
        if self.start_time is None:
            return "Unstarted"
        else:
            return f"{self.seconds - (time.time() - self.start_time):.0f}s"
    
    @property
    def started(self) -> bool:
        return self.start_time is not None

    def start(self):
        if self.start_time is None:
            self.start_time = time.time()
        return self

    def expired(self) -> bool:
        if self.start_time is None:
            return False
        else:
            return time.time() - self.start_time > self.seconds

    def reset(self):
        self.start_time = time.time()

class Interval:
    def __init__(self, seconds: float = 0.3):
        self.seconds = seconds
        self.start_time = time.time()
        self.last_wait_time = 0

    def wait(self):
        from .context import sleep
        delta = time.time() - self.start_time
        if delta < self.seconds:
            sleep(self.seconds - delta)
        self.last_wait_time = time.time() - self.start_time
        self.start_time = time.time()

    def reset(self):
        self.start_time = time.time()

def lf_path(path: str) -> str:
    standalone = os.path.join('kotonebot-resource', path)
    if os.path.exists(standalone):
        return standalone
    return str(resources.files('kaa.res') / path)

class Profiler:
    """
    性能分析器。对 `cProfile` 的简单封装。

    使用方法：
    ```python
    with Profiler('profile.prof'):
        # ...

    # 或者
    profiler = Profiler('profile.prof')
    profiler.begin()
    # ...
    profiler.end()
    ```
    """
    def __init__(self, file_path: str):

        self.profiler = cProfile.Profile()
        self.stats = None
        self.file_path = file_path

    def __enter__(self):
        self.profiler.enable()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.profiler.disable()
        self.stats = pstats.Stats(self.profiler)
        self.stats.dump_stats(self.file_path)

    def begin(self):
        self.__enter__()

    def end(self):
        self.__exit__(None, None, None)

    def snakeviz(self) -> bool:
        if self.stats is None:
            logger.warning("Profiler still running. Exit/End Profiler before run snakeviz.")
            return False
        try:
            from snakeviz import cli
            cli.main([os.path.abspath(self.file_path)])
            return True

        except ImportError:
            logger.warning("snakeviz is not installed")
            return False



import logging
from typing import Callable


from .. import R
from .loading import loading
from ..game_ui import toolbar_home
from kotonebot import device, image, action, cropped, until, sleep
from kotonebot.errors import UnrecoverableError

logger = logging.getLogger(__name__)


@action('检测是否位于首页')
def at_home() -> bool:
    with cropped(device, y1=0.7):
        return image.find(R.Daily.ButtonHomeCurrent) is not None

@action('检测是否位于日常商店页面')
def at_daily_shop() -> bool:
    icon = image.find(R.Daily.IconShopTitle)
    if icon is not None:
        return True
    else:
        # 调整默认购买数量的设置弹窗
        # [screenshots/contest/settings_popup.png]
        if image.find(R.Common.ButtonIconClose):
            device.click()
            sleep(1)
            return at_daily_shop()
        else:
            return False

@action('返回首页')
def goto_home():
    """
    从其他场景返回首页。

    前置条件：无 \n
    结束状态：位于首页
    """
    logger.info("Going home.")
    with cropped(device, y1=0.7):
        if home := toolbar_home():
            device.click(home)
            while loading():
                pass
        elif image.find(R.Common.ButtonHome):
            device.click()
        else:
            raise UnrecoverableError("Failed to go home.")
    image.expect_wait(R.Daily.ButtonHomeCurrent, timeout=20)

@action('前往商店页面')
def goto_shop():
    """
    从首页进入 ショップ。

    前置条件：无 \n
    结束状态：位于商店页面
    """
    logger.info("Going to shop.")
    if not at_home():
        goto_home()
    device.click(image.expect(R.Daily.ButtonShop))
    until(at_daily_shop, critical=True)


@action('测试颜色')
def test():
    from kotonebot import color
    while True:
        print(color.find_rgb('#ffffff', threshold=0.9999))
        print(image.find(R.Common.ButtonHome))

if __name__ == "__main__":
    import time
    test()



import time
from logging import getLogger

import cv2
import numpy as np

from kotonebot import image, device, debug, action, sleep
from kotonebot.backend.debug import result
from .. import R

logger = getLogger(__name__)

@action('检测加载页面', screenshot_mode='manual')
def loading() -> bool:
    """检测是否在场景加载页面"""
    img = device.screenshot()
    original_img = img.copy()
    # 二值化图片
    _, img = cv2.threshold(img, 127, 255, cv2.THRESH_BINARY)
    # 裁剪上面 20%
    img = img[:int(img.shape[0] * 0.2), :]
    # 判断图片中颜色数量是否 <= 2
    # https://stackoverflow.com/questions/56606294/count-number-of-unique-colours-in-image
    b,g,r = cv2.split(img)
    shiftet_im = b.astype(np.int64) + 1000 * (g.astype(np.int64) + 1) + 1000 * 1000 * (r.astype(np.int64) + 1)
    ret = len(np.unique(shiftet_im)) <= 2
    result('tasks.actions.loading', [img, original_img], f'result={ret}')
    return ret

@action('等待加载开始')
def wait_loading_start(timeout: float = 60):
    """等待加载开始"""
    start_time = time.time()
    while not loading():
        if time.time() - start_time > timeout:
            raise TimeoutError('加载超时')
        logger.debug('Not loading...')
        sleep(1)

@action('等待加载结束')
def wait_loading_end(timeout: float = 60):
    """等待加载结束"""
    start_time = time.time()
    while loading():
        if time.time() - start_time > timeout:
            raise TimeoutError('加载超时')
        # 检查网络错误
        if image.find(R.Common.TextNetworkError):
            device.click(image.expect(R.Common.ButtonRetry))
        logger.debug('Loading...')
        sleep(1)

if __name__ == '__main__':
    print(loading())
    input()



"""检测与跳过交流"""
import logging

from cv2.typing import MatLike

from kotonebot.backend.util import Countdown, Interval

from .. import R
from kotonebot import device, image, color, user, rect_expand, until, action, sleep, use_screenshot

logger = logging.getLogger(__name__)


@action('检查是否处于交流')
def is_at_commu():
    return image.find(R.Common.ButtonCommuFastforward) is not None

@action('跳过交流')
def skip_commu():
    device.click(image.expect_wait(R.Common.ButtonCommuSkip))

@action('检查未读交流', screenshot_mode='manual')
def handle_unread_commu(img: MatLike | None = None) -> bool:
    """
    检查当前是否处在未读交流，并自动跳过。

    :param img: 截图。
    :return: 是否跳过了交流。
    """
    ret = False
    logger.info('Check and skip commu')
    img = use_screenshot(img)
    skip_btn = image.find(R.Common.ButtonCommuFastforward)
    if skip_btn is None:
        logger.info('No fast forward button found. Not at a commu.')
        return ret
    
    ret = True
    logger.debug('Fast forward button found. Check commu')
    button_bg_rect = rect_expand(skip_btn.rect, 10, 10, 50, 10)
    def is_fastforwarding():
        nonlocal img
        assert img is not None
        colors = color.raw().dominant_color(img, 2, rect=button_bg_rect)
        RANGE = ((20, 65, 95), (180, 100, 100))
        return any(color.raw().in_range(c, RANGE) for c in colors)
    
    # 防止截图速度过快时，截图到了未加载完全的画面
    cd = Interval()
    hit = 0
    HIT_THRESHOLD = 2
    while True:
        if not is_fastforwarding():
            logger.debug("Unread commu hit %d/%d", hit, HIT_THRESHOLD)
            hit += 1
        else:
            hit = 0
            break
        if hit >= HIT_THRESHOLD:
            break
        cd.wait()
        img = device.screenshot()
    should_skip = hit >= HIT_THRESHOLD
    if not should_skip:
        logger.info('Fast forwarding. No action needed.')
        return False

    if should_skip:
        user.info('发现未读交流', images=[img])
        logger.debug('Not fast forwarding. Click fast forward button')
        device.click(skip_btn)
        sleep(0.7)
        if image.wait_for(R.Common.ButtonConfirm, timeout=5):
            logger.debug('Click confirm button')
            device.click()
    else:
        logger.info('Fast forwarding. No action needed.')
    logger.debug('Wait until not at commu')
    # TODO: 这里改用 while 循环，避免点击未生效的情况
    until(lambda: not is_at_commu(), interval=0.3)
    logger.info('Fast forward done')

    return ret


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=logging.INFO, format='[%(asctime)s] [%(levelname)s] [%(name)s] [%(funcName)s] [%(lineno)d] %(message)s')
    logger.setLevel(logging.DEBUG)
    print(is_at_commu())
    # rect = image.expect(R.Common.ButtonCommuFastforward).rect
    # print(rect)
    # rect = rect_expand(rect, 10, 10, 50, 10)
    # print(rect)
    # img = device.screenshot()
    # print(color.raw().dominant_color(img, 2, rect=rect))
    # skip_commu()
    # check_and_skip_commu()



import math
import time
import logging
from typing_extensions import deprecated
from itertools import cycle
from typing import Generic, Iterable, Literal, NamedTuple, Callable, Generator, TypeVar, ParamSpec, cast

import cv2
import numpy as np
from cv2.typing import MatLike


from .. import R
from . import loading
from ..common import conf
from .scenes import at_home
from kotonebot.errors import UnrecoverableError
from kotonebot.backend.context.context import use_screenshot
from .common import until_acquisition_clear, acquisitions, commut_event
from kotonebot.backend.util import AdaptiveWait, Countdown, crop, cropped
from kotonebot.backend.dispatch import DispatcherContext, SimpleDispatcher
from kotonebot import ocr, device, contains, image, regex, action, sleep, color, Rect, wait
from .non_lesson_actions import (
    enter_allowance, allowance_available, study_available, enter_study,
    is_rest_available, rest
)

class SkillCard(NamedTuple):
    available: bool
    rect: Rect

logger = logging.getLogger(__name__)

ActionType = None | Literal['lesson', 'rest']
@deprecated('OCR 方法效果不佳')
def enter_recommended_action_ocr(final_week: bool = False) -> ActionType:
    """
    在行动选择页面，执行推荐行动

    :param final_week: 是否是考试前复习周
    :return: 是否成功执行推荐行动
    """
    # 获取课程
    logger.debug("Waiting for recommended lesson...")
    with cropped(device, y1=0.00, y2=0.30):
        ret = ocr.wait_for(regex('ボーカル|ダンス|ビジュアル|休|体力'))
    logger.debug("ocr.wait_for: %s", ret)
    if ret is None:
        return None
    if not final_week:
        if "ボーカル" in ret.text:
            lesson_text = "Vo"
        elif "ダンス" in ret.text:
            lesson_text = "Da"
        elif "ビジュアル" in ret.text:
            lesson_text = "Vi"
        elif "休" in ret.text or "体力" in ret.text:
            rest()
            return 'rest'
        else:
            return None
        logger.info("Rec. lesson: %s", lesson_text)
        # 点击课程
        logger.debug("Try clicking lesson...")
        lesson_ret = ocr.expect(contains(lesson_text))
        device.double_click(lesson_ret.rect)
        return 'lesson'
    else:
        if "ボーカル" in ret.text:
            template = R.InPurodyuusu.ButtonFinalPracticeVocal
        elif "ダンス" in ret.text:
            template = R.InPurodyuusu.ButtonFinalPracticeDance
        elif "ビジュアル" in ret.text:
            template = R.InPurodyuusu.ButtonFinalPracticeVisual
        else:
            return None
        logger.debug("Try clicking lesson...")
        device.double_click(image.expect_wait(template))
        return 'lesson'

@action('执行推荐行动')
def handle_recommended_action(final_week: bool = False) -> ActionType:
    """
    在行动选择页面，执行推荐行动

    前置条件：位于行动选择页面\n
    结束状态：
        * `lesson`：练习场景，以及中间可能出现的加载、支援卡奖励、交流等
        * `rest`：休息动画。

    :param final_week: 是否是考试前复习周
    :return: 是否成功执行推荐行动
    """
    # 获取课程
    logger.debug("Getting recommended lesson...")
    with cropped(device, y1=0.00, y2=0.30):
        result = image.find_multi([
            R.InPurodyuusu.TextSenseiTipDance,
            R.InPurodyuusu.TextSenseiTipVocal,
            R.InPurodyuusu.TextSenseiTipVisual,
            R.InPurodyuusu.TextSenseiTipRest,
        ])
    logger.debug("image.find_multi: %s", result)
    if result is None:
        logger.debug("No recommended lesson found")
        return None
    if not final_week:
        if result.index == 0:
            lesson_text = regex("Da")
        elif result.index == 1:
            lesson_text = regex("Vo|V0|VO")
        elif result.index == 2:
            lesson_text = regex("Vi|V1|VI")
        elif result.index == 3:
            rest()
            return 'rest'
        else:
            return None
        logger.info("Rec. lesson: %s", lesson_text)
        # 点击课程
        logger.debug("Try clicking lesson...")
        lesson_ret = ocr.expect(lesson_text)
        device.double_click(lesson_ret.rect)
        return 'lesson'
    else:
        if result.index == 0:
            template = R.InPurodyuusu.ButtonFinalPracticeDance
        elif result.index == 1:
            template = R.InPurodyuusu.ButtonFinalPracticeVocal
        elif result.index == 2:
            template = R.InPurodyuusu.ButtonFinalPracticeVisual
        else:
            return None
        logger.debug("Try clicking lesson...")
        device.double_click(image.expect_wait(template))
        return 'lesson'

class CardDetectResult(NamedTuple):
    type: int
    """
    点击的卡片类型。

    0=第一张卡片，1=第二张卡片，2=第三张卡片，3=第四张卡片，10=SKIP。
    """
    score: float
    """总分数"""
    left_score: float
    """左边分数"""
    right_score: float
    """右边分数"""
    top_score: float
    """上边分数"""
    bottom_score: float
    """下边分数"""
    rect: Rect

def detect_recommended_card(
        card_count: int,
        threshold_predicate: Callable[[CardDetectResult], bool],
        *,
        img: MatLike | None = None,
    ):
    """
    识别推荐卡片

    前置条件：练习或考试中\n
    结束状态：-

    :param card_count: 卡片数量(2-4)
    :param threshold_predicate: 阈值判断函数
    :return: 执行结果。若返回 None，表示未识别到推荐卡片。
    """
    YELLOW_LOWER = np.array([20, 100, 100])
    YELLOW_UPPER = np.array([30, 255, 255])
    CARD_POSITIONS_1 = [
        # 格式：(x, y, w, h, return_value)
        (264, 883, 192, 252, 0)
    ]
    CARD_POSITIONS_2 = [
        (156, 883, 192, 252, 0),
        (372, 883, 192, 252, 1),
        # delta_x = 216, delta_x-width = 24
    ]
    CARD_POSITIONS_3 = [
        (47, 883, 192, 252, 0),  # 左卡片 (x, y, w, h)
        (264, 883, 192, 252, 1),  # 中卡片
        (481, 883, 192, 252, 2)   # 右卡片
        # delta_x = 217, delta_x-width = 25
    ]
    CARD_POSITIONS_4 = [
        (17, 883, 192, 252, 0),
        (182, 883, 192, 252, 1),
        (346, 883, 192, 252, 2),
        (511, 883, 192, 252, 3),
        # delta_x = 165, delta_x-width = -27
    ]
    SKIP_POSITION = (621, 739, 85, 85, 10)
    GLOW_EXTENSION = 15

    if card_count == 1:
        cards = CARD_POSITIONS_1
    elif card_count == 2:
        cards = CARD_POSITIONS_2
    elif card_count == 3:
        cards = CARD_POSITIONS_3
    elif card_count == 4:
        cards = CARD_POSITIONS_4
    else:
        raise ValueError(f"Unsupported card count: {card_count}")
    cards.append(SKIP_POSITION)

    
    image = use_screenshot(img)
    results: list[CardDetectResult] = []
    for x, y, w, h, return_value in cards:
        outer = (max(0, x - GLOW_EXTENSION), max(0, y - GLOW_EXTENSION))
        # 裁剪出检测区域
        glow_area = image[outer[1]:y + h + GLOW_EXTENSION, outer[0]:x + w + GLOW_EXTENSION]
        area_h = glow_area.shape[0]
        area_w = glow_area.shape[1]
        glow_area[GLOW_EXTENSION:area_h-GLOW_EXTENSION, GLOW_EXTENSION:area_w-GLOW_EXTENSION] = 0

        # 过滤出目标黄色
        glow_area = cv2.cvtColor(glow_area, cv2.COLOR_BGR2HSV)
        yellow_mask = cv2.inRange(glow_area, YELLOW_LOWER, YELLOW_UPPER)
        
        # 分割出每一边
        left_border = yellow_mask[:, 0:GLOW_EXTENSION]
        right_border = yellow_mask[:, area_w-GLOW_EXTENSION:area_w]
        top_border = yellow_mask[0:GLOW_EXTENSION, :]
        bottom_border = yellow_mask[area_h-GLOW_EXTENSION:area_h, :]
        y_border_pixels = area_h * GLOW_EXTENSION
        x_border_pixels = area_w * GLOW_EXTENSION

        # 计算每一边的分数
        left_score = np.count_nonzero(left_border) / y_border_pixels
        right_score = np.count_nonzero(right_border) / y_border_pixels
        top_score = np.count_nonzero(top_border) / x_border_pixels
        bottom_score = np.count_nonzero(bottom_border) / x_border_pixels

        result = (left_score + right_score + top_score + bottom_score) / 4
        results.append(CardDetectResult(
            return_value,
            result,
            left_score,
            right_score,
            top_score,
            bottom_score,
            (x, y, w, h)
        ))

    filtered_results = list(filter(threshold_predicate, results))
    if not filtered_results:
        max_result = max(results, key=lambda x: x.score)
        logger.info("Max card detect result (discarded): value=%d score=%.4f borders=(%.4f, %.4f, %.4f, %.4f)",
            max_result.type,
            max_result.score,
            max_result.left_score,
            max_result.right_score,
            max_result.top_score,
            max_result.bottom_score
        )
        return None
    filtered_results.sort(key=lambda x: x.score, reverse=True)
    logger.info("Max card detect result: value=%d score=%.4f borders=(%.4f, %.4f, %.4f, %.4f)",
        filtered_results[0].type,
        filtered_results[0].score,
        filtered_results[0].left_score,
        filtered_results[0].right_score,
        filtered_results[0].top_score,
        filtered_results[0].bottom_score
    )
    return filtered_results[0]

def handle_recommended_card(
        card_count: int, timeout: float = 7,
        threshold_predicate: Callable[[CardDetectResult], bool] = lambda _: True,
        *,
        img: MatLike | None = None,
    ):
    # cd = Countdown(seconds=timeout)
    # while not cd.expired():
    #     result = detect_recommended_card(card_count, threshold_predicate, img=img)
    #     if result is not None:
    #         device.double_click(result)
    #         return result
    #     sleep(np.random.uniform(0.01, 0.1))
    # return None

    result = detect_recommended_card(card_count, threshold_predicate, img=img)
    if result is not None:
        device.double_click(result)
        return result
    return None


@action('获取当前卡片数量', screenshot_mode='manual-inherit')
def skill_card_count(img: MatLike | None = None):
    """获取当前持有的技能卡数量"""
    device.click(0, 0)
    img = use_screenshot(img)
    img = crop(img, y1=0.83, y2=0.90)
    count = image.raw().count(img, R.InPurodyuusu.A)
    count += image.raw().count(img, R.InPurodyuusu.M)
    logger.info("Current skill card count: %d", count)
    return count


Yield = TypeVar('Yield')
Send = TypeVar('Send')
Return = TypeVar('Return')
P = ParamSpec('P')
class GeneratorWrapper(Iterable[Yield], Generic[P, Yield, Send, Return]):
    def __init__(
        self,
        generator_func: Callable[P, Generator[Yield, Send, Return]],
        *args: P.args,
        **kwargs: P.kwargs
    ):
        self.generator_func = generator_func
        self.generator = generator_func(*args, **kwargs)
        self.args = args
        self.kwargs = kwargs

    def __iter__(self):
        return self

    def __call__(self):
        return next(self.generator)

    def reset(self):
        self.generator = self.generator_func(*self.args, **self.kwargs)

    def loop(self) -> Return:
        while True:
            try:
                next(self.generator)
            except StopIteration as e:
                return cast(Return, e.value)

@action('获取当前卡牌信息', screenshot_mode='manual-inherit')
def obtain_cards(img: MatLike | None = None):
    img = use_screenshot(img)
    cards_rects = image.find_all_multi([
        R.InPurodyuusu.A,
        R.InPurodyuusu.M
    ])
    logger.info("Current cards: %s", len(cards_rects))
    cards = []
    for result in cards_rects:
        available = color.find_rgb('#7a7d7d', rect=result.rect) is None
        cards.append(SkillCard(available=available, rect=result.rect))
    return cards


@action('等待进入行动场景')
def until_action_scene(week_first: bool = False):
    """等待进入行动场景"""
    # 检测是否到行动页面
    while not image.find_multi([
        R.InPurodyuusu.TextPDiary, # 普通周
        R.InPurodyuusu.ButtonFinalPracticeDance # 离考试剩余一周
    ]):
        logger.info("Action scene not detected. Retry...")
        if acquisitions():
            continue
        if week_first and commut_event():
            continue
        sleep(0.2)
    else:
        logger.info("Now at action scene.")
        return 

@action('等待进入练习场景')
def until_practice_scene():
    """等待进入练习场景"""
    while image.find(R.InPurodyuusu.TextClearUntil) is None:
        until_acquisition_clear()

@action('等待进入考试场景')
def until_exam_scene():
    """等待进入考试场景"""
    # NOTE: is_exam_scene() 通过 OCR 剩余回合数判断是否处于考试场景。
    # 本来有可能会与练习场景混淆，
    # 但是在确定后续只是考试场景的情况下应该不会
    while ocr.find(regex("合格条件|三位以上")) is None and not is_exam_scene():
        until_acquisition_clear()

@action('执行练习', screenshot_mode='manual')
def practice():
    """
    执行练习
    
    前置条件：位于练习场景\n
    结束状态：各种奖励领取弹窗、加载画面等
    """
    logger.info("Practice started")

    def threshold_predicate(result: CardDetectResult):
        border_scores = (result.left_score, result.right_score, result.top_score, result.bottom_score)
        return (
            result.score >= 0.03
            # and len(list(filter(lambda x: x >= 0.01, border_scores))) >= 3
        )

    # 循环打出推荐卡，直到练习结束
    wait = cycle([0.1, 0.3, 0.5]) # 见下方解释
    tries = 1
    break_cd = Countdown(sec=3)
    while True:
        start_time = time.time()
        img = device.screenshot()
        if image.find(R.Common.ButtonIconCheckMark):
            logger.info("Confirmation dialog detected")
            device.click()
            sleep(4) # 等待卡片刷新
            continue

        card_count = skill_card_count(img)
        # TODO: 需要处理卡片数量为 0 的且当前回合未结束的情况
        if card_count > 0:
            inner_tries = 0
            # 为什么要内层循环：
            # 在 adb 截图模式下，如果把 skill_card_count 和 handle_recommended_card 
            # 放在同一循环中执行，循环一次大约需要 1.4~1.6s，
            # 游戏中推荐卡的发光周期约为 1s，也就是循环一次大约是 1.5 个发光周期。
            #
            # 假设发光强度符合规律 y=|sin t|（t=时间），
            # 如果第一次循环恰巧处于发光强度较弱时，
            # 那么经过 1.5T 后，依旧是处于发光强度较弱时，
            # 这样会导致每次循环都正好是在发光强度较弱时检测，
            # 每次都无法达到检测阈值。
            # 
            # 因此：
            # 1. 每次循环都等待随机时间
            # 2. 将 handle_recommended_card 单独放到内循环中执行，
            #    提高检测速度
            while True:
                result = handle_recommended_card(
                    card_count=card_count,
                    threshold_predicate=threshold_predicate,
                    img=img
                )
                if result:
                    break
                img = device.screenshot()
                inner_tries += 1
                if inner_tries >= 5:
                    break
            if result:
                logger.info("Handle recommended card success with %d tries", tries)
                sleep(4.5)
                tries = 0
            else:
                tries += 1
                end_time = time.time()
                logger.debug("Handle recommended card time: %.2f", end_time - start_time)
                delay = next(wait)
                logger.info("Tries: %d, Delay: %.2f", tries, delay)
                sleep(delay)
        
        if (
            card_count == 0
            and not image.find_multi([
                R.InPurodyuusu.TextClearUntil,
                R.InPurodyuusu.TextPerfectUntil
            ])
        ):
            if not break_cd.started:
                break_cd.start()
            if break_cd.expired():
                break
        else:
            break_cd.reset()

    # 结束动画
    logger.info("CLEAR/PERFECT not found. Practice finished.")
    (SimpleDispatcher('practice.end')
        .click(contains("上昇"), finish=True, log="Click to finish 上昇")
        .until(contains("審査基準"))
        .click('center')
    ).run()

@action('执行考试')
def exam(type: Literal['mid', 'final']):
    """
    执行考试
    
    前置条件：考试进行中场景（手牌可见）\n
    结束状态：考试结束交流/对话（TODO：截图）
    """
    logger.info("Exam started")

    def threshold_predicate(result: CardDetectResult):
        if type == 'final':
            return (
                result.score >= 0.4
                and result.left_score >= 0.2
                and result.right_score >= 0.2
                and result.top_score >= 0.2
                and result.bottom_score >= 0.2
            )
        else:
            return result.score >= 0.10
        # 关于上面阈值的解释：
        # 所有阈值均指卡片周围的“黄色度”，
        # score 指卡片四边的平均黄色度阈值，
        # left_score、right_score、top_score、bottom_score 指卡片每边的黄色度阈值

        # 为什么期中和期末考试阈值不一样：
        # 期末考试的场景为黄昏，背景中含有大量黄色，
        # 非常容易对推荐卡的检测造成干扰。
        # 解决方法是提高平均阈值的同时，为每一边都设置阈值。
        # 这样可以筛选出只有四边都包含黄色的发光卡片，
        # 而由夕阳背景造成的假发光卡片通常不会四边都包含黄色。

    wait = cycle([0.1, 0.3, 0.5])
    tries = 1
    no_card_cd = Countdown(sec=4)
    while True:
        start_time = time.time()
        img = device.screenshot()
        if image.find(R.Common.ButtonIconCheckMark):
            logger.info("Confirmation dialog detected")
            device.click()
            sleep(4) # 等待卡片刷新
            continue

        card_count = skill_card_count(img)
        if card_count == 0:
            # 处理本回合已无剩余手牌的情况
            # TODO: 使用模板匹配而不是 OCR，提升速度
            no_remaining_card = ocr.find(contains("0枚"), rect=R.InPurodyuusu.BoxNoSkillCard)
            if no_remaining_card:
                # TODO: HARD CODEDED
                SKIP_POSITION = (621, 739, 85, 85)
                device.click(SKIP_POSITION)
                no_card_cd.reset()
                continue
    
        if card_count > 0:
            inner_tries = 0
            while True:
                result = handle_recommended_card(
                    card_count=card_count,
                    threshold_predicate=threshold_predicate,
                    img=img
                )
                if result:
                    break
                img = device.screenshot()
                inner_tries += 1
                if inner_tries >= 5:
                    break
            if result:
                logger.info("Handle recommended card success with %d tries", tries)
                sleep(4.5)
                tries = 0
            else:
                tries += 1
                end_time = time.time()
                logger.debug("Handle recommended card time: %.2f", end_time - start_time)
                delay = next(wait)
                logger.info("Tries: %d, Delay: %.2f", tries, delay)
                sleep(delay)
        # 考试结束
        elif (
            card_count == 0
            and not ocr.find(contains('残りターン'), rect=R.InPurodyuusu.BoxExamTop)
            and image.find(R.Common.ButtonNext)
        ):
            # 点击“次へ”
            device.click()
            break

    if type == 'final':
        while ocr.wait_for(contains("メモリー"), timeout=7):
            device.click_center()

@action('考试结束流程')
def produce_end():
    """执行考试结束流程"""
    bottom = (int(device.screen_size[0] / 2), int(device.screen_size[1] * 0.9))
    # 1. 考试结束交流 [screenshots/produce/in_produce/final_exam_end_commu.png]
    # 2. 然后是，考试结束对话 [screenshots\produce_end\step2.jpg]
    # 3. MV
    # 4. 培育结束交流
    # 上面这些全部一直点就可以


    # 等待选择封面画面 [screenshots/produce_end/select_cover.jpg]
    # 次へ
    logger.info("Waiting for select cover screen...")
    aw = AdaptiveWait(timeout=60 * 5, max_interval=20)
    while not image.find(R.InPurodyuusu.ButtonNextNoIcon):
        aw()
        device.click(0, 0)
    # 选择封面
    logger.info("Use default cover.")
    sleep(3)
    logger.debug("Click next")
    device.click(image.expect_wait(R.InPurodyuusu.ButtonNextNoIcon))
    sleep(1)
    # 确认对话框 [screenshots/produce_end/select_cover_confirm.jpg]
    # 決定
    logger.debug("Click Confirm")
    device.click(image.expect_wait(R.Common.ButtonConfirm, threshold=0.8))
    sleep(1)
    # 上传图片，等待“生成”按钮
    # 注意网络可能会很慢，可能出现上传失败对话框
    logger.info("Waiting for cover uploading...")
    retry_count = 0
    MAX_RETRY_COUNT = 5
    while True:
        img = device.screenshot()
        # 处理上传失败
        if image.raw().find(img, R.InPurodyuusu.ButtonRetry):
            logger.info("Upload failed. Retry...")
            retry_count += 1
            if retry_count >= MAX_RETRY_COUNT:
                logger.info("Upload failed. Max retry count reached.")
                logger.info("Cancel upload.")
                device.click(image.expect_wait(R.InPurodyuusu.ButtonCancel))
                sleep(2)
                continue
            device.click()
        # 记忆封面保存失败提示
        elif image.raw().find(img, R.Common.ButtonClose):
            logger.info("Memory cover save failed. Click to close.")
            device.click()
        elif gen_btn := ocr.raw().find(img, contains("生成")):
            logger.info("Generate memory cover completed.")
            device.click(gen_btn)
            break
        else:
            device.click_center()
        sleep(2)
    # 后续动画
    logger.info("Waiting for memory generation animation completed...")
    while not image.find(R.InPurodyuusu.ButtonNextNoIcon):
        device.click_center()
        sleep(1)
    
    # 结算完毕
    # logger.info("Finalize")
    # # [screenshots/produce_end/end_next_1.jpg]
    # logger.debug("Click next 1")
    # device.click(image.expect_wait(R.InPurodyuusu.ButtonNextNoIcon))
    # sleep(1.3)
    # # [screenshots/produce_end/end_next_2.png]
    # logger.debug("Click next 2")
    # device.click(image.expect_wait(R.InPurodyuusu.ButtonNextNoIcon))
    # sleep(1.3)
    # # [screenshots/produce_end/end_next_3.png]
    # logger.debug("Click next 3")
    # device.click(image.expect_wait(R.InPurodyuusu.ButtonNextNoIcon))
    # sleep(1.3)
    # # [screenshots/produce_end/end_complete.png]
    # logger.debug("Click complete")
    # device.click(image.expect_wait(R.InPurodyuusu.ButtonComplete))
    # sleep(1.3)

    # 四个完成画面
    logger.info("Finalize")
    while True:
        # [screenshots/produce_end/end_next_1.jpg]
        # [screenshots/produce_end/end_next_2.png]
        # [screenshots/produce_end/end_next_3.png]
        if image.find(R.InPurodyuusu.ButtonNextNoIcon):
            logger.debug("Click next")
            device.click()
            wait(0.5, before='screenshot')
        # [screenshots/produce_end/end_complete.png]
        elif image.find(R.InPurodyuusu.ButtonComplete):
            logger.debug("Click complete")
            device.click(image.expect_wait(R.InPurodyuusu.ButtonComplete))
            wait(0.5, before='screenshot')
            break

    # 点击结束后可能还会弹出来：
    # 活动进度、关注提示
    # [screenshots/produce_end/end_activity.png]
    # [screenshots/produce_end/end_activity1.png]
    # [screenshots/produce_end/end_follow.png]
    # [screenshots/produce_end/end_level_up.png]
    while not at_home():
        if image.find(R.Common.ButtonIconClose):
            logger.info("Activity award claim dialog found. Click to close.")
            device.click()
        elif image.find(R.Common.ButtonNextNoIcon, colored=True):
            logger.debug("Click next")
            device.click()
        elif image.find(R.InPurodyuusu.ButtonCancel):
            logger.info("Follow producer dialog found. Click to close.")
            if conf().produce.follow_producer:
                logger.info("Follow producer")
                device.click(image.expect_wait(R.InPurodyuusu.ButtonFollowNoIcon))
            else:
                logger.info("Skip follow producer")
                device.click()
        else:
            device.click_center()
        sleep(1)
    logger.info("Produce completed.")

def week_normal(week_first: bool = False):
    until_action_scene(week_first)
    executed_action = handle_recommended_action()
    logger.info("Executed recommended action: %s", executed_action)
    # 推荐练习
    if executed_action == 'lesson':
        until_practice_scene()
        practice()
    # 推荐休息
    elif executed_action == 'rest':
        pass
    # 没有推荐行动
    elif executed_action is None:
        if allowance_available():
            enter_allowance()
        elif study_available():
            enter_study()
        elif is_rest_available():
            rest()
        else:
            raise ValueError("No action available.")
    until_action_scene()

def week_final_lesson():
    until_action_scene()
    if handle_recommended_action(final_week=True) != 'lesson':
        raise ValueError("Failed to enter recommended action on final week.")
    sleep(5)
    until_practice_scene()
    practice()

def week_mid_exam():
    logger.info("Week mid exam started.")
    logger.info("Wait for exam scene...")
    until_exam_scene()
    logger.info("Exam scene detected.")
    sleep(5)
    device.click_center()
    sleep(5)
    exam('mid')
    until_action_scene()

def week_final_exam():
    logger.info("Week final exam started.")
    logger.info("Wait for exam scene...")
    until_exam_scene()
    logger.info("Exam scene detected.")
    sleep(5)
    device.click_center()
    sleep(0.5)
    loading.wait_loading_end()
    exam('final')
    produce_end()

@action('执行 Regular 培育')
def hajime_regular(week: int = -1, start_from: int = 1):
    """
    「初」 Regular 模式

    :param week: 第几周，从1开始，-1表示全部
    :param start_from: 从第几周开始，从1开始。
    """
    weeks = [
        lambda: week_normal(True), # 1: Vo.レッスン、Da.レッスン、Vi.レッスン
        week_normal, # 2: 授業
        week_normal, # 3: Vo.レッスン、Da.レッスン、Vi.レッスン、授業
        week_normal, # 4: おでかけ、相談、活動支給
        week_final_lesson, # 5: 追い込みレッスン
        week_mid_exam, # 6: 中間試験
        week_normal, # 7: おでかけ、活動支給
        week_normal, # 8: 授業、活動支給
        week_normal, # 9: Vo.レッスン、Da.レッスン、Vi.レッスン
        week_normal, # 10: Vo.レッスン、Da.レッスン、Vi.レッスン、授業
        week_normal, # 11: おでかけ、相談、活動支給
        week_final_lesson, # 12: 追い込みレッスン
        week_final_exam, # 13: 最終試験
    ]
    if week == 0 or start_from == 0:
        until_action_scene(True)
    if week != -1:
        logger.info("Week %d started.", week)
        weeks[week - 1]()
    else:
        for i, w in enumerate(weeks[start_from-1:]):
            logger.info("Week %d started.", i + start_from)
            w()

@action('执行 PRO 培育')
def hajime_pro(week: int = -1, start_from: int = 1):
    """
    「初」 PRO 模式

    :param week: 第几周，从1开始，-1表示全部
    :param start_from: 从第几周开始，从1开始。
    """
    weeks = [
        lambda: week_normal(True), # 1
        week_normal, # 2
        week_normal, # 3
        week_normal, # 4
        week_normal, # 5
        week_final_lesson, # 6
        week_mid_exam, # 7
        week_normal, # 8
        week_normal, # 9
        week_normal, # 10
        week_normal, # 11
        week_normal, # 12
        week_normal, # 13
        week_normal, # 14
        week_final_lesson, # 15
        week_final_exam, # 16
    ]
    if week != -1:
        logger.info("Week %d started.", week)
        weeks[week - 1]()
    else:
        for i, w in enumerate(weeks[start_from-1:]):
            logger.info("Week %d started.", i + start_from)
            w()

@action('是否在考试场景')
def is_exam_scene():
    """是否在考试场景"""
    return ocr.find(contains('残りターン'), rect=R.InPurodyuusu.BoxExamTop) is not None

ProduceStage = Literal[
    'action', # 行动场景
    'practice-ongoing', # 练习场景
    'exam-start', # 考试开始确认页面
    'exam-ongoing', # 考试进行中
    'exam-end', # 考试结束
    'unknown', # 未知场景
]

@action('检测当前培育场景', dispatcher=True)
def detect_regular_produce_scene(ctx: DispatcherContext) -> ProduceStage:
    """
    判断当前是培育的什么阶段，并开始 Regular 培育。

    前置条件：培育中的任意场景\n
    结束状态：游戏主页面\n
    """
    logger.info("Detecting current produce stage...")
    
    # 行动场景
    texts = ocr.ocr()
    if (
        image.find_multi([
            R.InPurodyuusu.TextPDiary, # 普通周
            R.InPurodyuusu.ButtonFinalPracticeDance # 离考试剩余一周
        ]) 
    ):
        logger.info("Detection result: At action scene.")
        ctx.finish()
        return 'action'
    elif texts.where(regex('CLEARまで|PERFECTまで')):
        logger.info("Detection result: At practice ongoing.")
        ctx.finish()
        return 'practice-ongoing'
    elif is_exam_scene():
        logger.info("Detection result: At exam scene.")
        ctx.finish()
        return 'exam-ongoing'
    elif texts.where(regex('合格条件|三位以上')):
        logger.info("Detection result: At exam start.")
        ctx.finish()
        return 'exam-start'
    else:
        if acquisitions():
            return 'unknown'
        if commut_event():
            return 'unknown'
        return 'unknown'

@action('开始 Regular 培育')
def hajime_regular_from_stage(stage: ProduceStage):
    """
    开始 Regular 培育。
    """
    if stage == 'action':
        texts = ocr.ocr(rect=R.InPurodyuusu.BoxWeeksUntilExam)
        # 提取周数
        remaining_week = texts.squash().replace('ó', '6').numbers()
        if not remaining_week:
            raise UnrecoverableError("Failed to detect week.")
        # 判断阶段
        if texts.where(contains('中間')):
            week = 6 - remaining_week[0]
            hajime_regular(start_from=week)
        elif texts.where(contains('最終')):
            week = 13 - remaining_week[0]
            hajime_regular(start_from=week)
        else:
            raise UnrecoverableError("Failed to detect produce stage.")
    elif stage == 'exam-start':
        device.click_center()
        until_exam_scene()
        exam()
    elif stage == 'exam-ongoing':
        # TODO: 应该直接调用 week_final_exam 而不是再写一次
        logger.info("Exam ongoing. Start exam.")
        exam()
        result = ocr.expect_wait(contains('中間|最終'))
        if '中間' in result.text:
            return hajime_regular_from_stage(detect_regular_produce_scene())
        elif '最終' in result.text:
            produce_end()
        else:
            raise UnrecoverableError("Failed to detect produce stage.")
    elif stage == 'practice-ongoing':
        # TODO: 应该直接调用 week_final_exam 而不是再写一次
        logger.info("Practice ongoing. Start practice.")
        practice()
        return hajime_regular_from_stage(detect_regular_produce_scene())
    else:
        raise UnrecoverableError(f'Cannot resume produce REGULAR from stage "{stage}".')

@action('继续 Regular 培育')
def resume_regular_produce():
    """
    继续 Regular 培育。
    """
    hajime_regular_from_stage(detect_regular_produce_scene())

if __name__ == '__main__':
    from logging import getLogger

    logging.basicConfig(level=logging.INFO, format='[%(asctime)s] [%(levelname)s] [%(name)s] %(message)s')
    getLogger('kotonebot').setLevel(logging.DEBUG)
    getLogger(__name__).setLevel(logging.DEBUG)
    import os
    from datetime import datetime
    os.makedirs('logs', exist_ok=True)
    log_filename = datetime.now().strftime('logs/task-%y-%m-%d-%H-%M-%S.log')
    file_handler = logging.FileHandler(log_filename, encoding='utf-8')
    file_handler.setFormatter(logging.Formatter('[%(asctime)s] [%(levelname)s] [%(name)s] %(message)s'))
    logging.getLogger().addHandler(file_handler)

    from kotonebot.backend.util import Profiler
    from kotonebot.backend.context import init_context, manual_context
    init_context()
    manual_context().begin()

    # hajime_regular(start_from=1)
    
    # pf = Profiler('profiler')
    # pf.begin()
    # # do_produce(conf().produce.idols[0], 'pro')
    # practice()
    # hajime_pro(start_from=16)
    # pf.end()
    # pf.snakeviz()


    # while True:
    #     cards = obtain_cards()
    #     print(cards)
    #     sleep(1)


    # practice()
    # week_final_exam()
    # exam('final')
    # produce_end()


    # hajime_pro(start_from=16)
    # exam('mid')
    stage = (detect_regular_produce_scene())
    hajime_regular_from_stage(stage)

    # click_recommended_card(card_count=skill_card_count())
    # exam('mid')

    # hajime_regular(start_from=7)

    # import cv2
    # while True:
    #     img = device.screenshot()
    #     cv2.imshow('123', img)
    #     cv2.waitKey(1)



"""竞赛"""
import logging
from gettext import gettext as _

from kotonebot.backend.dispatch import SimpleDispatcher

from . import R
from .common import conf
from .actions.scenes import at_home, goto_home
from .actions.loading import wait_loading_end
from kotonebot import device, image, ocr, color, action, task, user, rect_expand, sleep, contains

logger = logging.getLogger(__name__)

@action('前往竞赛页面')
def goto_contest() -> bool:
    """
    前置条件：位于首页 \n
    结束状态：位于竞赛界面，且已经点击了各种奖励领取提示

    :return: 是否存在未完成的挑战
    """
    # TODO: 优化这一部分，加入区域检测，提高速度
    device.click(image.expect(R.Common.ButtonContest))
    ocr.expect_wait(contains('CONTEST'))
    btn_contest = ocr.expect_wait(contains('コンテスト'))
    has_ongoing_contest = image.find(R.Daily.TextContestLastOngoing) is not None
    device.click(btn_contest)
    if not has_ongoing_contest:
        while not image.find(R.Daily.ButtonContestRanking):
            # [screenshots/contest/acquire1.png]
            # [screenshots/contest/acquire2.png]
            device.click_center()
            sleep(1)
        # [screenshots/contest/main.png]
    else:
        image.expect_wait(R.Daily.ButtonContestChallengeStart)
    return has_ongoing_contest

@action('选择并挑战')
def pick_and_contest(has_ongoing_contest: bool = False) -> bool:
    """
    选择并挑战

    前置条件：位于竞赛界面 \n
    结束状态：位于竞赛界面

    :param has_ongoing_contest: 是否有中断未完成的挑战
    :return: 如果返回假，说明今天挑战次数已经用完了
    """
    # 判断是否有中断未完成的挑战
    # [screenshots/contest/ongoing.png]
    if not has_ongoing_contest:
        image.expect_wait(R.Daily.ButtonContestRanking)
        sleep(1) # 等待动画
        logger.info('Randomly pick a contestant and start challenge.')
        # 随机选一个对手 [screenshots/contest/main.png]
        logger.debug('Clicking on contestant.')
        contestant = image.wait_for(R.Daily.TextContestOverallStats, timeout=2)
        if contestant is None:
            logger.info('No contestant found. Today\'s challenge points used up.')
            return False
        device.click(contestant)
        # 挑战开始 [screenshots/contest/start1.png]
        logger.debug('Clicking on start button.')
        device.click(image.expect_wait(R.Daily.ButtonContestStart))
    sleep(1)
    # 进入挑战页面 [screenshots/contest/contest1.png]
    # [screenshots/contest/contest2.png]
    while not image.find(R.Daily.ButtonContestChallengeStart):
        # 记忆未编成 [screenshots/contest/no_memo.png]
        if image.find(R.Daily.TextContestNoMemory):
            logger.debug('Memory not set. Using auto-compilation.')
            user.warning('竞赛未编成', _('记忆未编成。将使用自动编成。'), once=True)
            device.click(image.expect(R.Daily.ButtonContestChallenge))
        logger.debug('Waiting for challenge start screen.')
    # 勾选跳过所有
    if image.find(R.Common.CheckboxUnchecked):
        logger.debug('Checking skip all.')
        device.click()
        sleep(0.5)
    # 点击 SKIP
    logger.debug('Clicking on SKIP.')
    device.click(image.expect(R.Daily.ButtonIconSkip, colored=True, transparent=True, threshold=0.999))
    while not image.wait_for(R.Common.ButtonNextNoIcon, timeout=2):
        device.click_center()
        logger.debug('Waiting for the result.')
    # [screenshots/contest/after_contest1.png]
    # 点击 次へ [screenshots/contest/after_contest2.png]
    logger.debug('Challenge finished. Clicking on next.')
    device.click()
    # 点击 終了 [screenshots/contest/after_contest3.png]
    logger.debug('Clicking on end.')
    device.click(image.expect_wait(R.Common.ButtonEnd))
    # 可能出现的奖励弹窗 [screenshots/contest/after_contest4.png]
    sleep(1)
    if image.find(R.Common.ButtonClose):
        logger.debug('Clicking on close.')
        device.click()
    # 等待返回竞赛界面
    wait_loading_end()
    image.expect_wait(R.Daily.ButtonContestRanking)
    logger.info('Challenge finished.')
    return True

@task('竞赛')
def contest():
    """"""
    if not conf().contest.enabled:
        logger.info('Contest is disabled.')
        return
    logger.info('Contest started.')
    if not at_home():
        goto_home()
    sleep(0.3)
    btn_contest = image.expect(R.Common.ButtonContest)
    notification_dot = rect_expand(btn_contest.rect, top=35, right=35)
    if not color.find_rgb('#ff104a', rect=notification_dot):
        logger.info('No action needed.')
        return
    has_ongoing_contest = goto_contest()
    while pick_and_contest(has_ongoing_contest):
        sleep(1.3)
        has_ongoing_contest = False
    goto_home()
    logger.info('Contest all finished.')
    

if __name__ == '__main__':
    import logging
    logging.basicConfig(level=logging.INFO, format='[%(asctime)s] [%(levelname)s] [%(name)s] [%(funcName)s] [%(lineno)d] %(message)s')
    logging.getLogger('kotonebot').setLevel(logging.DEBUG)
    logger.setLevel(logging.DEBUG)
    
    contest()


import logging
from typing import Callable, cast, Literal, overload
from typing_extensions import override, deprecated

import numpy as np
import cv2
from cv2.typing import MatLike
from adbutils import AdbClient, adb
from adbutils._device import AdbDevice

from kotonebot.backend.core import HintBox
from kotonebot.backend.util import Rect, is_rect
from .protocol import ClickableObjectProtocol, Commandable, Touchable, Screenshotable


logger = logging.getLogger(__name__)

# class AdbDevice(DeviceABC):
#     def __init__(self, device: Device) -> None:
#         super().__init__()
#         self.device = device

#     @override
#     def launch_app(self, package_name: str) -> None:
#         self.device.shell(f"monkey -p {package_name} 1")
    

#     @override
#     def swipe(self, x1: int, y1: int, x2: int, y2: int, duration: float|None = None) -> None:

#         if duration is not None:
#             logger.warning("Swipe duration is not supported with AdbDevice. Ignoring duration.")
#         self.device.shell(f"input touchscreen swipe {x1} {y1} {x2} {y2}")

#     @override
#     def screenshot(self) -> MatLike:
#         if self.screenshot_hook_before is not None:
#             logger.debug("execute screenshot hook before")
#             img = self.screenshot_hook_before()
#             if img is not None:
#                 logger.debug("screenshot hook before returned image")
#                 return img
#         img = self.screenshot_raw()
#         if self.screenshot_hook_after is not None:
#             img = self.screenshot_hook_after(img)
#         return img

#     @override
#     def screenshot_raw(self) -> MatLike:
#         return cv2.cvtColor(np.array(self.device.screenshot()), cv2.COLOR_RGB2BGR)

class HookContextManager:
    def __init__(self, device: 'Device', func: Callable[[MatLike], MatLike]):
        self.device = device
        self.func = func
        self.old_func = device.screenshot_hook_after

    def __enter__(self):
        self.device.screenshot_hook_after = self.func
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.device.screenshot_hook_after = self.old_func


class PinContextManager:
    def __init__(self, device: 'Device'):
        self.device = device
        self.old_hook = device.screenshot_hook_before
        self.memo = None

    def __hook(self) -> MatLike:
        if self.memo is None:
            self.memo = self.device.screenshot_raw()
        return self.memo

    def __enter__(self):
        self.device.screenshot_hook_before = self.__hook
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        self.device.screenshot_hook_before = self.old_hook

    def update(self) -> None:
        """
        更新记住的截图
        """
        self.memo = self.device.screenshot_raw()

class Device:
    def __init__(self, adb_connection: AdbDevice | None = None) -> None:
        self._adb: AdbDevice | None = adb_connection

        self.screenshot_hook_after: Callable[[MatLike], MatLike] | None = None
        """截图后调用的函数"""
        self.screenshot_hook_before: Callable[[], MatLike | None] | None = None
        """截图前调用的函数。返回修改后的截图。"""
        self.click_hooks_before: list[Callable[[int, int], tuple[int, int]]] = []
        """点击前调用的函数。返回修改后的点击坐标。"""
        self.last_find: Rect | ClickableObjectProtocol | None = None
        """上次 image 对象或 ocr 对象的寻找结果"""
        self.orientation: Literal['portrait', 'landscape'] = 'portrait'
        """
        设备当前方向。默认为竖屏。

        横屏时为 'landscape'，竖屏时为 'portrait'。
        """

        self._command: Commandable
        self._touch: Touchable
        self._screenshot: Screenshotable
    
    @property
    def adb(self) -> AdbDevice:
        if self._adb is None:
            raise ValueError("AdbClient is not connected")
        return self._adb

    @adb.setter
    def adb(self, value: AdbDevice) -> None:
        self._adb = value

    def launch_app(self, package_name: str) -> None:
        """
        根据包名启动 app
        """
        self._command.launch_app(package_name)
    
    @overload
    def click(self) -> None:
        """
        点击上次 `image` 对象或 `ocr` 对象的寻找结果（仅包括返回单个结果的函数）。
        （不包括 `image.raw()` 和 `ocr.raw()` 的结果。）

        如果没有上次寻找结果或上次寻找结果为空，会抛出异常 ValueError。
        """
        ...

    @overload
    def click(self, x: int, y: int) -> None:
        """
        点击屏幕上的某个点
        """
        ...

    @overload
    def click(self, hint_box: HintBox) -> None:
        """
        点击屏幕上的某个矩形区域
        """
        ...

    @overload
    def click(self, rect: Rect) -> None:
        """
        从屏幕上的某个矩形区域随机选择一个点并点击
        """
        ...

    @overload
    def click(self, clickable: ClickableObjectProtocol) -> None:

        """
        点击屏幕上的某个可点击对象
        """
        ...

    def click(self, *args, **kwargs) -> None:
        arg1 = args[0] if len(args) > 0 else None
        arg2 = args[1] if len(args) > 1 else None
        if arg1 is None:
            self.__click_last()
        elif isinstance(arg1, HintBox):
            self.__click_hint_box(arg1)
        elif is_rect(arg1):
            self.__click_rect(arg1)
        elif isinstance(arg1, int) and isinstance(arg2, int):
            self.__click_point(arg1, arg2)
        elif isinstance(arg1, ClickableObjectProtocol):
            self.__click_clickable(arg1)
        else:
            raise ValueError(f"Invalid arguments: {arg1}, {arg2}")

    def __click_last(self) -> None:
        if self.last_find is None:
            raise ValueError("No last find result. Make sure you are not calling the 'raw' functions.")
        self.click(self.last_find)

    def __click_rect(self, rect: Rect) -> None:
        # 从矩形中心的 60% 内部随机选择一点
        x = rect[0] + rect[2] // 2 + np.random.randint(-int(rect[2] * 0.3), int(rect[2] * 0.3))
        y = rect[1] + rect[3] // 2 + np.random.randint(-int(rect[3] * 0.3), int(rect[3] * 0.3))
        x = int(x)
        y = int(y)
        self.click(x, y)

    def __click_point(self, x: int, y: int) -> None:
        for hook in self.click_hooks_before:
            logger.debug(f"Executing click hook before: ({x}, {y})")
            x, y = hook(x, y)
            logger.debug(f"Click hook before result: ({x}, {y})")
        logger.debug(f"Click: {x}, {y}")
        self._touch.click(x, y)

    def __click_clickable(self, clickable: ClickableObjectProtocol) -> None:
        self.click(clickable.rect)

    def __click_hint_box(self, hint_box: HintBox) -> None:
        self.click(hint_box.rect)
    

    def click_center(self) -> None:
        """
        点击屏幕中心。
        
        此方法会受到 `self.orientation` 的影响。
        调用前确保 `orientation` 属性与设备方向一致，
        否则点击位置会不正确。
        """
        x, y = self.screen_size[0] // 2, self.screen_size[1] // 2
        self.click(x, y)
    
    @overload
    def double_click(self, x: int, y: int, interval: float = 0.25) -> None:
        """
        双击屏幕上的某个点
        """
        ...

    @overload
    def double_click(self, rect: Rect, interval: float = 0.25) -> None:
        """
        双击屏幕上的某个矩形区域
        """
        ...
    
    @overload
    def double_click(self, clickable: ClickableObjectProtocol, interval: float = 0.25) -> None:
        """
        双击屏幕上的某个可点击对象
        """
        ...
    
    def double_click(self, *args, **kwargs) -> None:
        from kotonebot import sleep
        arg0 = args[0]
        if is_rect(arg0) or isinstance(arg0, ClickableObjectProtocol):
            rect = arg0
            interval = kwargs.get('interval', 0.25)
            self.click(rect)
            sleep(interval)
            self.click(rect)
        else:
            x = args[0]
            y = args[1]
            interval = kwargs.get('interval', 0.25)
            self.click(x, y)
            sleep(interval)
            self.click(x, y)

    def swipe(self, x1: int, y1: int, x2: int, y2: int, duration: float|None = None) -> None:
        """
        滑动屏幕
        """
        self._touch.swipe(x1, y1, x2, y2, duration)

    def swipe_scaled(self, x1: float, y1: float, x2: float, y2: float, duration: float|None = None) -> None:
        """
        滑动屏幕，参数为屏幕坐标的百分比

        :param x1: 起始点 x 坐标百分比。范围 [0, 1]
        :param y1: 起始点 y 坐标百分比。范围 [0, 1]
        :param x2: 结束点 x 坐标百分比。范围 [0, 1]
        :param y2: 结束点 y 坐标百分比。范围 [0, 1]
        :param duration: 滑动持续时间，单位秒。None 表示使用默认值。
        """
        w, h = self.screen_size
        self.swipe(int(w * x1), int(h * y1), int(w * x2), int(h * y2), duration)
    
    def screenshot(self) -> MatLike:
        """
        截图
        """
        if self.screenshot_hook_before is not None:
            logger.debug("execute screenshot hook before")
            img = self.screenshot_hook_before()
            if img is not None:
                logger.debug("screenshot hook before returned image")
                return img
        img = self.screenshot_raw()
        if self.screenshot_hook_after is not None:
            img = self.screenshot_hook_after(img)
        return img

    def screenshot_raw(self) -> MatLike:
        """
        截图，不调用任何 Hook。
        """
        return self._screenshot.screenshot()

    def hook(self, func: Callable[[MatLike], MatLike]) -> HookContextManager:
        """
        注册 Hook，在截图前将会调用此函数，对截图进行处理
        """
        return HookContextManager(self, func)

    @deprecated('改用 @task/@action 装饰器中的 screenshot_mode 参数')
    def pinned(self) -> PinContextManager:
        """
        记住下次截图结果，并将截图调整为手动挡。
        之后截图都会返回记住的数据，节省重复截图时间。

        调用返回对象中的 PinContextManager.update() 可以立刻更新记住的截图。
        """
        return PinContextManager(self)

    @property
    def screen_size(self) -> tuple[int, int]:
        """
        屏幕尺寸。格式为 `(width, height)`。
        
        **注意**： 此属性返回的分辨率会随设备方向变化。
        如果 `self.orientation` 为 `landscape`，则返回的分辨率是横屏下的分辨率，
        否则返回竖屏下的分辨率。

        `self.orientation` 属性默认为竖屏。如果需要自动检测，
        调用 `self.detect_orientation()` 方法。
        如果已知方向，也可以直接设置 `self.orientation` 属性。
        """
        return self._screenshot.screen_size

    def current_package(self) -> str | None:
        """
        获取前台 APP 的包名。

        :return: 前台 APP 的包名。如果获取失败，则返回 None。
        :exception: 如果设备不支持此功能，则抛出 NotImplementedError。
        """
        ret = self._command.current_package()
        logger.debug("current_package: %s", ret)
        return ret

    def detect_orientation(self) -> Literal['portrait', 'landscape'] | None:
        """
        检测当前设备方向并设置 `self.orientation` 属性。

        :return: 检测到的方向，如果无法检测到则返回 None。
        """
        return self._screenshot.detect_orientation()

        
if __name__ == "__main__":
    from kotonebot.client.implements.adb import AdbImpl
    from kotonebot.client.implements.adb_raw import AdbRawImpl
    from .implements.uiautomator2 import UiAutomator2Impl
    print("server version:", adb.server_version())
    adb.connect("127.0.0.1:5555")
    print("devices:", adb.device_list())
    d = adb.device_list()[-1]
    d.shell("dumpsys activity top | grep ACTIVITY | tail -n 1")
    dd = Device(d)
    adb_imp = AdbRawImpl(dd)
    dd._command = adb_imp
    dd._touch = adb_imp
    dd._screenshot = adb_imp
    # dd._screenshot = MinicapScreenshotImpl(dd)
    # dd._screenshot = UiAutomator2Impl(dd)

    # 实时展示画面
    import cv2
    import numpy as np
    import time
    last_time = time.time()
    while True:
        start_time = time.time()
        img = dd.screenshot()
        # 50% 缩放
        img = cv2.resize(img, (img.shape[1] // 2, img.shape[0] // 2))
        
        # 计算帧间隔
        interval = start_time - last_time
        fps = 1 / interval if interval > 0 else 0
        last_time = start_time
        
        # 获取当前时间和帧率信息
        current_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        fps_text = f"FPS: {fps:.1f} {interval*1000:.1f}ms"
        
        # 在图像上绘制信息
        font = cv2.FONT_HERSHEY_SIMPLEX
        cv2.putText(img, current_time, (10, 30), font, 0.5, (0, 0, 255), 1, cv2.LINE_AA)
        cv2.putText(img, fps_text, (10, 60), font, 0.5, (0, 0, 255), 1, cv2.LINE_AA)
        
        cv2.imshow("screen", img)
        cv2.waitKey(1)


import os
import time
import subprocess
import struct
from threading import Thread
from functools import cached_property
from typing_extensions import override

import cv2
import numpy as np
from cv2.typing import MatLike
from adbutils._utils import adb_path

from .adb import AdbImpl
from ..device import Device
from kotonebot import logging

logger = logging.getLogger(__name__)

WAIT_TIMEOUT = 10
SCRIPT = """#!/bin/sh
while true; do
    screencap
    sleep 0.3
done
"""

class AdbRawImpl(AdbImpl):
    def __init__(self, device: Device):
        super().__init__(device)
        self.__worker: Thread | None = None
        self.__process: subprocess.Popen | None = None
        self.__data: MatLike | None = None

    def __worker_thread(self) -> None:
        with open('screenshot.sh', 'w', encoding='utf-8', newline='\n') as f:
            f.write(SCRIPT)
        self.adb.push('screenshot.sh', '/data/local/tmp/screenshot.sh')
        self.adb.shell(f'chmod 755 /data/local/tmp/screenshot.sh')
        os.remove('screenshot.sh')

        cmd = fr'{adb_path()} -s {self.adb.serial} exec-out "sh /data/local/tmp/screenshot.sh"'
        self.__process = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        
        while self.__process.poll() is None:
            if self.__process.stdout is None:
                logger.error("Failed to get stdout from process")
                continue
            
            # 解析 header  
            # https://stackoverflow.com/questions/22034959/what-format-does-adb-screencap-sdcard-screenshot-raw-produce-without-p-f
            if self.__api_level >= 26:
                metadata = self.__process.stdout.read(16)
                w, h, p, c = struct.unpack('<IIII', metadata)
                # w=width, h=height, p=pixel_format, c=color_space
                # 详见：https://android.googlesource.com/platform/frameworks/base/+/26a2b97dbe48ee45e9ae70110714048f2f360f97/cmds/screencap/screencap.cpp#209
            else:
                metadata = self.__process.stdout.read(12)
                w, h, p = struct.unpack('<III', metadata)
            if p == 1: # PixelFormat.RGBA_8888
                channel = 4
            else:
                raise ValueError(f"Unsupported pixel format: {p}")
            data_size = w * h * channel

            if (data_size < 100 * 100 * 4) or (data_size > 3000 * 3000 * 4):
                raise ValueError(f"Invaild data_size: {w}x{h}.")

            # 读取图像数据
            # logger.verbose(f"receiving image data: {w}x{h} {data_size} bytes")
            image_data = self.__process.stdout.read(data_size)
            if not isinstance(image_data, bytes) or len(image_data) != data_size:
                logger.error(f"Failed to read image data, expected {data_size} bytes but got {len(image_data) if isinstance(image_data, bytes) else 'non-bytes'}")
                continue
                
            np_data = np.frombuffer(image_data, np.uint8)
            np_data = np_data.reshape(h, w, channel)
            self.__data = cv2.cvtColor(np_data, cv2.COLOR_RGBA2BGR)

    @cached_property
    def __api_level(self) -> int:
        try:
            output = self.adb.shell("getprop ro.build.version.sdk")
            assert isinstance(output, str)
            return int(output.strip())
        except Exception as e:
            logger.error(f"Failed to get API level: {e}")
            return 0

    @override
    def screenshot(self) -> MatLike:
        if not self.__worker:
            self.__worker = Thread(target=self.__worker_thread, daemon=True)
            self.__worker.start()
        
        start_time = time.time()
        while self.__data is None:
            time.sleep(0.01)
            if time.time() - start_time > WAIT_TIMEOUT:
                raise TimeoutError("Failed to get screenshot from device.")
        logger.verbose(f"adb raw screenshot wait time: {time.time() - start_time:.4f}s")
        data = self.__data
        self.__data = None
        return data


import time
import uuid
import logging
import inspect
from logging import Logger
from types import CodeType
from dataclasses import dataclass
from typing import Annotated, Any, Callable, Concatenate, Sequence, TypeVar, ParamSpec, Literal, Protocol, cast
from typing_extensions import Self

from dataclasses import dataclass

from kotonebot.backend.ocr import StringMatchFunction
from kotonebot.backend.util import Rect, is_rect

from .core import Image

logger = logging.getLogger(__name__)
P = ParamSpec('P')
R = TypeVar('R')
ThenAction = Literal['click', 'log']
DoAction = Literal['click']
# TODO: 需要找个地方统一管理这些属性名
ATTR_DISPATCHER_MARK = '__kb_dispatcher_mark'
ATTR_ORIGINAL_FUNC = '_kb_inner'


class DispatchFunc: pass

wrapper_to_func: dict[Callable, Callable] = {}

class DispatcherContext:
    def __init__(self):
        self.finished: bool = False
        self._first_run: bool = True
    
    def finish(self):
        """标记已完成 dispatcher 循环。循环将在下次条件检测时退出。"""
        self.finished = True

    def expand(self, func: Annotated[Callable[[], Any], DispatchFunc], ignore_finish: bool = True):
        """
        调用其他 dispatcher 函数。

        使用 `expand` 和直接调用的区别是：
        * 直接调用会执行 while 循环，直到满足结束条件
        * 而使用 `expand` 则只会执行一次。效果类似于将目标函数里的代码直接复制粘贴过来。
        """
        # 获取原始函数
        original_func = func
        while not getattr(original_func, ATTR_DISPATCHER_MARK, False):
            original_func = getattr(original_func, ATTR_ORIGINAL_FUNC)
        original_func = getattr(original_func, ATTR_ORIGINAL_FUNC)

        if not original_func:
            raise ValueError(f'{repr(func)} is not a dispatcher function.')
        elif not callable(original_func):
            raise ValueError(f'{repr(original_func)} is not callable.')
        original_func = cast(Callable[[DispatcherContext], Any], original_func)

        old_finished = self.finished
        ret = original_func(self)
        if ignore_finish:
            self.finished = old_finished
        return ret

    @property
    def beginning(self) -> bool:
        """是否为第一次运行"""
        return self._first_run
    
    @property
    def finishing(self) -> bool:
        """是否即将结束运行"""
        return self.finished

def dispatcher(
        func: Callable[Concatenate[DispatcherContext, P], R],
        *,
        fragment: bool = False
    ) -> Annotated[Callable[P, R], DispatchFunc]:
    """
    注意：\n
    此装饰器必须在应用 @action/@task 装饰器后再应用，且 `screenshot_mode='manual'` 参数必须设置。
    或者也可以使用 @action/@task 装饰器中的 `dispatcher=True` 参数，
    那么就没有上面两个要求了。

    :param fragment:
        片段模式，默认不启用。
        启用后，被装饰函数将会只执行依次，
        而不会一直循环到 ctx.finish() 被调用。
    """
    def wrapper(*args: P.args, **kwargs: P.kwargs):
        ctx = DispatcherContext()
        while not ctx.finished:
            from kotonebot import device
            device.screenshot()
            ret = func(ctx, *args, **kwargs)
            ctx._first_run = False
        return ret
    def fragment_wrapper(*args: P.args, **kwargs: P.kwargs):
        ctx = DispatcherContext()
        from kotonebot import device
        device.screenshot()
        return func(ctx, *args, **kwargs)
    setattr(wrapper, ATTR_ORIGINAL_FUNC, func)
    setattr(fragment_wrapper, ATTR_ORIGINAL_FUNC, func)
    setattr(wrapper, ATTR_DISPATCHER_MARK, True)
    setattr(fragment_wrapper, ATTR_DISPATCHER_MARK, True)
    wrapper_to_func[wrapper] = func
    if fragment:
        return fragment_wrapper

    else:
        return wrapper

@dataclass
class ClickParams:
    finish: bool = False
    log: str | None = None

class ClickCenter:
    def __init__(self, sd: 'SimpleDispatcher', target: Image | str | StringMatchFunction | Literal['center'], *, params: ClickParams = ClickParams()):
        self.target = target
        self.params = params
        self.sd = sd

    def __call__(self):
        from kotonebot import device
        if self.params.log:
            self.sd.logger.info(self.params.log)
        device.click_center()
        if self.params.finish:
            self.sd.finished = True

class ClickImage:
    def __init__(self, sd: 'SimpleDispatcher', image: Image, *, params: ClickParams = ClickParams()):
        self.image = image
        self.params = params
        self.sd = sd

    def __call__(self):
        from kotonebot import device, image
        if image.find(self.image):
            if self.params.log:
                self.sd.logger.info(self.params.log)
            device.click()
            if self.params.finish:
                self.sd.finished = True

class ClickImageAny:
    def __init__(self, sd: 'SimpleDispatcher', images: list[Image], params: ClickParams = ClickParams()):
        self.images = images
        self.params = params
        self.sd = sd
    
    def __call__(self):
        from kotonebot import device, image
        if image.find_multi(self.images):
            if self.params.log:
                self.sd.logger.info(self.params.log)
            device.click()
            if self.params.finish:
                self.sd.finished = True

class ClickText:
    def __init__(
            self,
            sd: 'SimpleDispatcher',
            text: str | StringMatchFunction,
            params: ClickParams = ClickParams()
        ):
        self.text = text
        self.params = params
        self.sd = sd

    def __call__(self):
        from kotonebot import device, ocr
        if ocr.find(self.text):
            if self.params.log:
                self.sd.logger.info(self.params.log)
            device.click()
            if self.params.finish:
                self.sd.finished = True

class ClickRect:
    def __init__(self, sd: 'SimpleDispatcher', rect: Rect, *, params: ClickParams = ClickParams()):
        self.rect = rect
        self.params = params
        self.sd = sd

    def __call__(self):
        from kotonebot import device
        if device.click(self.rect):
            if self.params.log:
                self.sd.logger.info(self.params.log)
            if self.params.finish:
                self.sd.finished = True

class UntilText:
    def __init__(
            self,
            sd: 'SimpleDispatcher',
            text: str | StringMatchFunction,
            *,
            rect: Rect | None = None,
            result: Any | None = None
        ):
        self.text = text
        self.sd = sd
        self.rect = rect
        self.result = result

    def __call__(self):
        from kotonebot import ocr
        if ocr.find(self.text, rect=self.rect):
            self.sd.finished = True
            self.sd.result = self.result

class UntilImage:
    def __init__(
            self,
            sd: 'SimpleDispatcher',
            image: Image,
            *,
            rect: Rect | None = None,
            result: Any | None = None
        ):
        self.image = image
        self.sd = sd
        self.rect = rect
        self.result = result

    def __call__(self):
        from kotonebot import image
        if self.rect:
            logger.warning(f'UntilImage with rect is deprecated. Use UntilText instead.')
        if image.find(self.image):
            self.sd.finished = True
            self.sd.result = self.result

class SimpleDispatcher:
    def __init__(self, name: str, *, min_interval: float = 0.3):
        self.name = name
        self.logger = logging.getLogger(f'SimpleDispatcher of {name}')
        self.blocks: list[Callable] = []
        self.finished: bool = False
        self.result: Any | None = None
        self.min_interval = min_interval
        self.timeout_value: float | None = None
        self.timeout_critical: bool = False
        self.__last_run_time: float = 0

    def click(
        self,
        target: Image | StringMatchFunction | Literal['center'] | Rect,
        *,
        finish: bool = False,
        log: str | None = None
    ):
        params = ClickParams(finish=finish, log=log)
        if isinstance(target, Image):
            self.blocks.append(ClickImage(self, target, params=params))
        elif is_rect(target):
            self.blocks.append(ClickRect(self, target, params=params))
        elif callable(target):
            self.blocks.append(ClickText(self, target, params=params))
        elif target == 'center':
            self.blocks.append(ClickCenter(self, target='center', params=params))
        else:
            raise ValueError(f'Invalid target: {target}')
        return self

    def click_any(
        self,
        target: list[Image],
        *,
        finish: bool = False,
        log: str | None = None
    ):
        params = ClickParams(finish=finish, log=log)
        self.blocks.append(ClickImageAny(self, target, params))
        return self

    def until(
        self,
        text: StringMatchFunction | Image,
        *,
        rect: Rect | None = None,
        result: Any | None = None
    ):
        if isinstance(text, Image):
            self.blocks.append(UntilImage(self, text, rect=rect, result=result))
        else:
            self.blocks.append(UntilText(self, text, rect=rect, result=result))
        return self

    def timeout(self, timeout: float, *, critical: bool = False, result: Any | None = None):
        self.timeout_value = timeout
        self.timeout_critical = critical
        self.timeout_result = result
        return self

    def run(self):
        from kotonebot import device, sleep
        while True:
            logger.debug(f'Running dispatcher "{self.name}"')
            time_delta = time.time() - self.__last_run_time
            if time_delta < self.min_interval:
                sleep(self.min_interval - time_delta)
            for block in self.blocks:
                block()
            if self.finished:
                break
            self.__last_run_time = time.time()
            if self.timeout_value and time.time() - self.__last_run_time > self.timeout_value:
                if self.timeout_critical:
                    raise TimeoutError(f'Dispatcher "{self.name}" timed out.')
                else:
                    self.logger.warning(f'Dispatcher "{self.name}" timed out.')
                    self.result = self.timeout_result
                    break
            device.screenshot()
        return self.result

if __name__ == '__main__':
    from .context.task_action import action
    from .context import init_context
    init_context()
    @action('inner', dispatcher=True)
    def inner(ctx: DispatcherContext):
        print('inner')
        ctx.finish()

    @action('test', dispatcher=True)
    def test(ctx: DispatcherContext):
        print('test')
        inner()
        ctx.finish()
    test()
